#include "GreedySolver.hpp"
#include "Game.hpp"
#include "Map.hpp"
#include <random>

using std::vector;

#define VERBOSE 0

GreedySolver::GreedySolver() :
    m_nextMove((uint32_t)-1)
{
}

GreedySolver::~GreedySolver()
{
}

uint32_t GreedySolver::solve(const Game& game)
{
    if (m_nextMove != (uint32_t)-1)
    {
        uint32_t nextMove = m_nextMove;
        m_nextMove = (uint32_t)-1;
        return nextMove;
    }

    auto& map = game.m_map;
    uint32_t numSites = map.m_numSites;
    uint32_t numRivers = map.m_numRivers;
    uint32_t numMines = map.m_numMines;
    uint32_t punter = game.m_punter;

    uint32_t numComponents = 0;
    vector<uint32_t> siteComponents(numSites);
    for (uint32_t iSite = 0; iSite < numSites; iSite++)
    {
        siteComponents[iSite] = (uint32_t)-1;
    }

    vector<uint32_t> stack;
    for (uint32_t iSite = 0; iSite < numSites; iSite++)
    {
        if (siteComponents[iSite] != (uint32_t)-1)
        {
            continue;
        }

        stack.push_back(iSite);
        while (!stack.empty())
        {
            uint32_t iCurSite = stack.back();
            stack.pop_back();

            if (siteComponents[iCurSite] != (uint32_t)-1)
            {
                continue;
            }
            siteComponents[iCurSite] = numComponents;

            uint32_t low = map.m_siteRiverBounds[iCurSite].m_low;
            uint32_t high = map.m_siteRiverBounds[iCurSite].m_high;
            for (uint32_t iSiteRiver = low; iSiteRiver < high; iSiteRiver++)
            {
                auto& siteRiver = map.m_siteRivers[iSiteRiver];
                if (game.m_claims[siteRiver.m_river] != punter)
                {
                    continue;
                }

                uint32_t target = siteRiver.m_target;
                if (siteComponents[target] == (uint32_t)-1)
                {
                    stack.push_back(target);
                }
            }
        }

        numComponents++;
    }

#if VERBOSE
    printf("numComponents = %" PRIu32 "\n", numComponents);
#endif

    uint32_t numMineComponents = 0;
    vector<uint32_t> componentMineComponents(numComponents);
    for (uint32_t iComponent = 0; iComponent < numComponents; iComponent++)
    {
        componentMineComponents[iComponent] = (uint32_t)-1;
    }
    for (uint32_t iMine = 0; iMine < numMines; iMine++)
    {
        uint32_t iSite = map.m_mines[iMine];
        uint32_t iComponent = siteComponents[iSite];
        if (componentMineComponents[iComponent] != (uint32_t)-1)
        {
            continue;
        }

        componentMineComponents[iComponent] = numMineComponents;
        numMineComponents++;
    }

#if VERBOSE
    printf("numMineComponents = %" PRIu32 "\n", numMineComponents);
#endif

    vector<uint32_t> scoreTable(numMineComponents * numComponents);

    vector<uint32_t> queueCurrent;
    vector<uint32_t> queueNext;
    vector<uint8_t> visited(numSites);
    for (uint32_t iMine = 0; iMine < numMines; iMine++)
    {
        uint32_t iSite = map.m_mines[iMine];
        uint32_t iComponent = siteComponents[iSite];
        uint32_t iMineComponent = componentMineComponents[iComponent];

        for (uint32_t iSite = 0; iSite < numSites; iSite++)
        {
            visited[iSite] = 0;
        }

        uint32_t dist = 0;
        queueNext.push_back(iSite);
        while (!queueNext.empty())
        {
            queueCurrent.swap(queueNext);
            for (uint32_t iCurSite : queueCurrent)
            {
                if (visited[iCurSite])
                {
                    continue;
                }
                visited[iCurSite] = 1;
                uint32_t iCurComponent = siteComponents[iCurSite];
                if (iCurComponent != iComponent)
                {
                    scoreTable[iMineComponent * numComponents + iCurComponent]
                        += dist * dist;
                }

                uint32_t low = map.m_siteRiverBounds[iCurSite].m_low;
                uint32_t high = map.m_siteRiverBounds[iCurSite].m_high;
                for (uint32_t iSiteRiver = low; iSiteRiver < high; iSiteRiver++)
                {
                    auto& siteRiver = map.m_siteRivers[iSiteRiver];
                    uint32_t target = siteRiver.m_target;
                    if (!visited[target])
                    {
                        queueNext.push_back(target);
                    }
                }
            }
            queueCurrent.clear();
            dist++;
        }
    }

    vector<uint32_t> siteLiberties(numSites);
    for (uint32_t iRiver = 0; iRiver < numRivers; iRiver++)
    {
        if (game.m_claims[iRiver] != (uint32_t)-1)
        {
            continue;
        }

        auto& river = map.m_rivers[iRiver];
        siteLiberties[river.m_source]++;
        siteLiberties[river.m_target]++;
    }

    uint32_t bestMove = (uint32_t)-1;
    uint32_t bestScore = 0;
    uint32_t bestLiberties = 0;
    uint32_t bestCount = 0;

    std::default_random_engine gen{std::random_device()()};

    for (uint32_t iRiver = 0; iRiver < numRivers; iRiver++)
    {
        if (game.m_claims[iRiver] != (uint32_t)-1)
        {
            continue;
        }

        auto& river = map.m_rivers[iRiver];
        uint32_t iSourceComponent = siteComponents[river.m_source];
        uint32_t iTargetComponent = siteComponents[river.m_target];
        uint32_t iSourceMineComponent = componentMineComponents[iSourceComponent];
        uint32_t iTargetMineComponent = componentMineComponents[iTargetComponent];

        uint32_t score = 0;
        if (iSourceMineComponent != (uint32_t)-1)
        {
            score += scoreTable[iSourceMineComponent * numComponents + iTargetComponent];
        }
        if (iTargetMineComponent != (uint32_t)-1)
        {
            score += scoreTable[iTargetMineComponent * numComponents + iSourceComponent];
        }

        uint32_t liberties = 0;
        liberties += siteLiberties[river.m_source];
        liberties += siteLiberties[river.m_target];

#if VERBOSE
        printf("%4" PRIu32 ": %" PRIu32 " (%" PRIu32 " liberties)\n",
               iRiver,
               score,
               liberties);
#endif

        if (bestMove == (uint32_t)-1 ||
            score > bestScore ||
            (score == bestScore && liberties > bestLiberties))
        {
            bestMove = iRiver;
            bestScore = score;
            bestLiberties = liberties;
            bestCount = 1;
        }
        else if (score == bestScore &&
                 liberties == bestLiberties)
        {
            bestCount++;
            std::uniform_int_distribution<uint32_t> dist(0, bestCount - 1);
            if (dist(gen) == 0)
            {
                bestMove = iRiver;
            }
        }
    }

    if (bestMove != (uint32_t)-1)
    {
#if VERBOSE
        printf("Best:\n");
        printf("%4" PRIu32 ": %" PRIu32 " (%" PRIu32 " liberties)\n",
               bestMove,
               bestScore,
               bestLiberties);
#endif
        return bestMove;
    }




    {
        auto& map = game.m_map;
        uint32_t numRivers = map.m_numRivers;
        uint32_t numUnclaimed = 0;
        for (uint32_t iRiver = 0; iRiver < numRivers; iRiver++)
        {
            if (game.m_claims[iRiver] == (uint32_t)-1)
            {
                numUnclaimed++;
            }
        }

        std::default_random_engine gen{std::random_device()()};
        std::uniform_int_distribution<uint32_t> dist(0, numUnclaimed - 1);
        uint32_t iUnclaimed = dist(gen);

        numUnclaimed = 0;
        for (uint32_t iRiver = 0; iRiver < numRivers; iRiver++)
        {
            if (game.m_claims[iRiver] == (uint32_t)-1)
            {
                if (numUnclaimed == iUnclaimed)
                {
                    return iRiver;
                }
                numUnclaimed++;
            }
        }
    }
    
    return (uint32_t)-1;
}

void GreedySolver::getFutures(const Game& game,
                              vector<SolverFuture>* pFutures)
{
    if (game.m_punter != 0)
    {
        return;
    }

    m_nextMove = solve(game);

    if (m_nextMove == (uint32_t)-1)
    {
        return;
    }

    auto& river = game.m_map.m_rivers[m_nextMove];
    if (game.m_map.m_isMine[river.m_source] &&
        !game.m_map.m_isMine[river.m_target])
    {
        pFutures->emplace_back(river.m_source, river.m_target);
    }
    else if (game.m_map.m_isMine[river.m_target] &&
             !game.m_map.m_isMine[river.m_source])
    {
        pFutures->emplace_back(river.m_target, river.m_source);
    }
}

void GreedySolver::serialize(vector<uint32_t>* pData) const
{
    pData->push_back(m_nextMove);
}

bool GreedySolver::deserialize(const vector<uint32_t>& data,
                               size_t& pos)
{
    size_t size = data.size();
    if (pos + 1 > size)
    {
        return false;
    }

    m_nextMove = data[pos++];
    return true;
}
