#include "MapGraphics.hpp"
#include "Graphics.hpp"
#include "TimeUtils.hpp"

using std::string;
using std::vector;

MapGraphics::MapGraphics(const string& title,
                         const Map& map,
                         const vector<uint32_t>* pClaims,
                         uint32_t punter) :
    m_title(title),
    m_map(map),
    m_pClaims(),
    m_punter((uint32_t)-1),
    m_graphics(),
    m_destroyRequested(false),
    m_destroyed(false)
{
    setClaims(pClaims, punter);
}

MapGraphics::~MapGraphics()
{
}

void MapGraphics::show()
{
    string msg;
    if (!create(&msg))
    {
        printf("%s\n", msg.c_str());
        return;
    }

    while (true)
    {
        while (Graphics::update(false))
        {
            if (m_destroyed)
            {
                break;
            }
        }

        if (m_destroyed)
        {
            break;
        }

        sleepMS(10);
    }
}

bool MapGraphics::create(string* pMsg)
{
    constexpr uint32_t colors[] =
    {
        0x1f77b4,
        0xaec7e8,
        0xff7f0e,
        0xffbb78,
        0x2ca02c,
        0x98df8a,
        0xd62728,
        0xff9896,
        0x9467bd,
        0xc5b0d5,
        0x8c564b,
        0xc49c94,
        0xe377c2,
        0xf7b6d2,
        0x7f7f7f,
        0xc7c7c7,
        0xbcbd22,
        0xdbdb8d,
        0x17becf,
        0x9edae5
    };
    constexpr uint32_t numColors = sizeof(colors)/sizeof(colors[0]);
    
    m_destroyRequested = false;
    m_destroyed = false;

    auto onDestroy = [this]()
    {
        m_destroyed = true;
    };
    m_graphics.setOnDestroy(onDestroy);

    auto onPaint = [this]()
    {
        auto& gr = m_graphics;

        gr.beginPaint(true);

        uint32_t width = gr.getWidth();
        uint32_t height = gr.getHeight();

        if (width > 0 && height > 0)
        {
            double windowRatio = (double)width / (double)height;

            double mapWidth = m_map.m_maxCoord.m_x - m_map.m_minCoord.m_x;
            double mapHeight = m_map.m_maxCoord.m_y - m_map.m_minCoord.m_y;

            double windowWidth = 0.0;
            double windowHeight = 0.0;

            if (mapWidth == 0.0 && mapHeight == 0.0)
            {
                windowWidth = windowRatio;
                windowHeight = 1.0;
            }
            else if (mapWidth == 0.0)
            {
                windowHeight = mapHeight;
                windowWidth = mapHeight * windowRatio;
            }
            else if (mapHeight == 0.0)
            {
                windowWidth = mapWidth;
                windowHeight = mapWidth / windowRatio;
            }
            else
            {
                double mapRatio = mapWidth / mapHeight;
                if (mapRatio > windowRatio)
                {
                    windowWidth = mapWidth;
                    windowHeight = mapWidth / windowRatio;
                }
                else
                {
                    windowHeight = mapHeight;
                    windowWidth = mapHeight * windowRatio;
                }
            }

            windowWidth *= 1.05;
            windowHeight *= 1.05;

            double midX = (m_map.m_minCoord.m_x + m_map.m_maxCoord.m_x) / 2.0;
            double midY = (m_map.m_minCoord.m_y + m_map.m_maxCoord.m_y) / 2.0;

            vector<std::pair<int32_t, int32_t>> screenCoords(m_map.m_numSites);
            for (uint32_t iSite = 0; iSite < m_map.m_numSites; iSite++)
            {
                double dx = m_map.m_siteCoords[iSite].m_x;
                double dy = m_map.m_siteCoords[iSite].m_y;

                int32_t x = (int32_t)((dx - midX) / windowWidth * (double)width) + ((int32_t)width / 2);
                int32_t y = (int32_t)((dy - midY) / windowHeight * (double)height) + ((int32_t)height / 2);

                screenCoords[iSite].first = x;
                screenCoords[iSite].second = y;
            }

            gr.fillRect(0,
                        0,
                        width,
                        height,
                        0x000000);

            for (uint32_t iRiver = 0; iRiver < m_map.m_numRivers; iRiver++)
            {
                auto& river = m_map.m_rivers[iRiver];
                int32_t x1 = screenCoords[river.m_source].first;
                int32_t y1 = screenCoords[river.m_source].second;
                int32_t x2 = screenCoords[river.m_target].first;
                int32_t y2 = screenCoords[river.m_target].second;

                uint32_t color = 0x0000ff;
                uint32_t lineWidth = 1;

                if (m_pClaims)
                {
                    uint32_t punter = (*m_pClaims)[iRiver];
                    if (punter != (uint32_t)-1)
                    {
                        if (punter == m_punter)
                        {
                            color = 0xffffff;
                            lineWidth = 5;
                        }
                        else
                        {
                            color = colors[punter % numColors];
                            lineWidth = 3;
                        }
                    }
                }

                gr.setLineWidth(lineWidth);
                gr.drawLine(x1, y1, x2, y2, color);
            }

            int32_t siteCircleSize = 3;
            int32_t siteCircleHalf = siteCircleSize / 2;
            for (uint32_t iSite = 0; iSite < m_map.m_numSites; iSite++)
            {
                int32_t x = screenCoords[iSite].first;
                int32_t y = screenCoords[iSite].second;
                gr.fillCircle(x - siteCircleHalf,
                              y - siteCircleHalf,
                              siteCircleSize,
                              siteCircleSize,
                              0xffffff);
            }

            int32_t mineCircleSize = 9;
            int32_t mineCircleHalf = mineCircleSize / 2;
            for (uint32_t iMine = 0; iMine < m_map.m_numMines; iMine++)
            {
                int32_t x = screenCoords[m_map.m_mines[iMine]].first;
                int32_t y = screenCoords[m_map.m_mines[iMine]].second;
                gr.fillCircle(x - mineCircleHalf,
                              y - mineCircleHalf,
                              mineCircleSize,
                              mineCircleSize,
                              0xff0000);
            }
        }

        gr.endPaint();
    };
    m_graphics.setOnPaint(onPaint);

    auto onKey = [this](Key key)
    {
        switch (key)
        {
        case Key::Q:
        case Key::Esc:
        {
            if (!m_destroyRequested && !m_destroyed)
            {
                m_destroyRequested = true;
                m_graphics.destroy();
            }
            break;
        }
        default: ;
        }
    };
    m_graphics.setOnKey(onKey);

    if (!m_graphics.create(1000,
                           1000,
                           m_title.c_str(),
                           pMsg))
    {
        return false;
    }

    return true;
}

void MapGraphics::onClaims(const vector<uint32_t>* pClaims,
                           uint32_t punter)
{
    setClaims(pClaims, punter);
    m_graphics.requestPaint();
}

void MapGraphics::destroy()
{
    if (!m_destroyRequested && !m_destroyed)
    {
        m_destroyRequested = true;
        m_graphics.destroy();
    }
}

void MapGraphics::setClaims(const vector<uint32_t>* pClaims,
                            uint32_t punter)
{
    if (pClaims)
    {
        m_pClaims.reset(new vector<uint32_t>(*pClaims));
    }
    else
    {
        m_pClaims.reset();
    }
    m_punter = punter;
}
